/* ----------------------------------------------------------------------
 * FILE: lexer.c
 * PACKAGE: as504 - Siemens SAB-C504 Assembler.
 *
 * DESCRIPTION:
 *	This file contains the lexical tokenizer for the assembler.
 *	Since yacc is being used the lexer is called yylex().
 *
 *	In order to produce a listing, some record of the users
 *	source line must be kept. This is done by adding
 *	get_ch(), and unget_ch() routine which returns/ungets a character
 *	but also places information into a secret array.
 *
 *	When a newline is encountered the text line is returned as
 *	an attribute on the '\n' character.
 *
 * LICENCE:
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 * REVISION HISTORY:
 *	Jan. 19, 1990 - Created. (Ken Stauffer)
 *	Jan. 12, 2005 - Fixed bug: reading binary values (Freek van Walderveen)
 *  Mar. 23, 2005 - Added support for includefiles and some other changes
 *                  (Freek van Walderveen)
 *
 * AUTHOR:
 *	Original file written by Ken Stauffer (University of Calgary).
 *	January, 1990.
 *
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>
#define YYSTYPE union ystack
#include "as504.h"

void error(cs,a1,a2,a3,a4,a5,a6); // in as504.y

extern union ystack yylval;
extern int pass, debug;
extern char *asmfile;
/* Include stack */
extern FILE *fp_stack[];
extern int linenumber_stack[];
extern char *filename_stack[];
extern int if_sp_stack[];
extern int include_sp;
/* IF stack */
extern int if_sp;

struct symbol *looksym();
struct opcode *lookop();
//char *malloc();
int lineno;
FILE *fin;

static char line[100],*lineptr=line;

/* ----------------------------------------------------------------------
 * get_ch:
 *	Get a character from fin, place char in line[]
 */

char get_ch()
{
	register int c;

	c = getc(fin);
	if( c != EOF && lineptr - line < sizeof(line) )
		*lineptr++ = c;
	return(c);
}

/* ----------------------------------------------------------------------
 * unget_ch:
 *	Unget a character and move lineptr back by one.
 */

void unget_ch(int c)
{
	ungetc(c,fin);
	if( lineptr > line )
		lineptr--;
}

/* ----------------------------------------------------------------------
 * yylex:
 *	The tokens themselves are returned via return(token)
 *
 *	Some tokens have attributes. These attributes are returned
 *	by setting a global variable yylval:
 *
 *		yylval.value
 *			numbers (any base)
 *			strings (in pass 1).
 *			bit positions .0, .1, .2, ...
 *
 *		yylval.str
 *			strings (in pass 2).
 *			'\n' (both passes).
 *
 *		yylval.sym
 *			User defined symbols.
 *
 *		yylval.op
 *			Reserved keyword (opcode/directive/misc.)
 *
 *		No other fields in yylval are used by yylex().
 * 
 *		Characters that do not have an attribute do
 *		not set anything in the yylval variable.
 *
 */



int yylex(void)
{
	static int nl_flag=0;	/* sync. error messages and the cur. line */
	register int c;
	char buf[256];		/* temporary buffer */
	char *p;		/* general pointer */
	struct symbol *sym;
	struct opcode *op;

  char quot_type;
	int octal=0,hex=0,decimal=0,binary=0;
  static int eol_at_eof = 0;
	register long value = 0;

	if( nl_flag ) {
		nl_flag = 0;
		lineno++;
	}

for(;;) {
	c = get_ch();
	switch(c) {
	case EOF:
    if (!eol_at_eof) {
      eol_at_eof++;
      return '\n';
    }
    eol_at_eof = 0;
    if (include_sp == 0) {
      if (if_sp > 0) {
        error("IF directive not ended at end of file.");
      }
      /* end of main file */
      return(EOF);
    } else {
      /* end of some included file */
      
      /* check for stack underflow */
      if (include_sp < 0) {
	      error("include stackpointer < 0, this is a bug, please report it");
      }
      
      include_sp--;
      if (if_sp > if_sp_stack[include_sp]) {
        error("IF directive not ended at end of included file.");
      }
      
      /* close included file */
      (void) fclose(fin);

      if (debug) {
        printf("Popping includefile ``%s''\n", asmfile);
      }
      
      /* pop off the line number, if_sp and file pointer */
      fin = fp_stack[include_sp];
      lineno = linenumber_stack[include_sp] + 1;
      if_sp = if_sp_stack[include_sp];
      asmfile = (char *)realloc(asmfile,(unsigned)strlen(filename_stack[include_sp]) + 1);
      strcpy(asmfile, filename_stack[include_sp]);
      free(filename_stack[include_sp]);
    }
	case ' ':
	case '\t':
  case '\r':
		break;

	case '\n':
		nl_flag = 1;
		yylval.str = line;
		*lineptr = '\0';
		lineptr = line;
		return('\n');

	case ';':
		while((c=get_ch()) != EOF && c!='\n');
		nl_flag= 1;
		yylval.str = line;
		*lineptr = '\0';
		lineptr = line;
		return(c);

#ifdef WINDOWS_QUOTING
  case '\'':
#endif
  case '"':
    quot_type = c;
		p = buf;
		while((c=get_ch()) != EOF && c!=quot_type && c!='\n') {
			if( c == '\\' ) {
				switch(c=get_ch()) {
				case 'n': c = '\n'; break;
				case 'r': c = '\r'; break;
				case 't': c = '\t'; break;
				case 'b': c = '\b'; break;
				case '"': c = '"'; break;
        case '\'': c = '\''; break;
				case '\\': c = '\\'; break;
				default:
				  error("Invalid escape character: \\%c",c);
				  break;
				}
			}
			if( p-buf<sizeof(buf)-1 ) {
				*p++ = c;
			} else {
			   error("String constant longer than %d bytes",
					sizeof(buf));
			}
		}
    *p = '\0';
    if( c == '\n' || c == EOF ) {
      error("String terminated improperly.");
      unget_ch(c);
    }
    if( (p = malloc(strlen(buf)+1)) == NULL ) {
      error("Cannot allocate %d bytes",strlen(buf)+1);
    }
    strcpy(p,buf);
    yylval.str = p;
		return(STRING);

	case '.':
		if( (c=get_ch())>='0' && c<='7' ) {
			yylval.value = c-'0';
			return(BITPOS);
		}
		unget_ch(c);
		return('.');

#ifndef WINDOWS_QUOTING
	case '\'':
		c = get_ch();
		if( c=='\\' ) {
			switch(c=get_ch()) {
			case 'n': c = '\n'; break;
			case 'r': c = '\r'; break;
			case 't': c = '\t'; break;
			case 'b': c = '\b'; break;
			case '0': c = 0; break;
			case 'o': c = 0; break;
			case 'O': c = 0; break;
			case '\\': c = '\\'; break;
			case '\'': c = '\''; break;
			default:
				error("Invalid escape character: \\%c",c);
			}
		}
		if( get_ch() != '\'' )
			error("Missing quote in character constant");
		yylval.value = c;
		return(VALUE);
#endif
    
  case '|':
    return OR;
  case '^':
    return XOR;
  case '&':
    return AND;
  case '~':
    return BITNOT;
  
  case '$':
    c = get_ch();
    if (!isxdigit(c)) {
      unget_ch(c);
      return '$';
    }
    /* else: hexadecimal number */
    hex++;
  case '%':
    if (!hex) {
      c = get_ch();
      if (c == ' ' || c == '\t') {
        return MOD;
      } else if (c != '0' && c != '1') {
        return '%'; // generates syntax error, there could be a nicer way te achieve this
      } /* else: binary number */
      binary++;
    }
	case '0':	/* parse a number			*/
	case '1':	/* could be followed by a:		*/
	case '2':	/*	'b','B' - Binary		*/
	case '3':	/*	'h','H' - Hex			*/
	case '4':	/*	'd','D' - Decimal		*/
	case '5':	/*	'o','O' - Octal			*/
	case '6':	/* *** Numbers must start with a digit	*/
	case '7':	/* Numbers could be also preceeded by:  */
	case '8':	/*	0x	- Hex,	0b     - binary */
	case '9':	/*	0	- Octal			*/

		p = buf;
		do {
			if( p-buf<sizeof(buf)-1 )
				*p++ = c;
			c = get_ch();
		} while( c=='H' || c=='h' || c=='O' || c=='o' ||
				c=='x' || c=='X' || isxdigit(c) );
		unget_ch(c);
		*p = '\0';

    if (!binary && !hex) {
		  /* Check any preceeding chars */
		  if( buf[0]=='0' && (buf[1]=='x' || buf[1]=='X') ) {
				  hex++;
				  buf[1] = '0';
		  } else {
			  if( buf[0]=='0' &&
			    (buf[1]=='b' || buf[1]=='B') &&
			    *(p-1) != 'h' && *(p-1) != 'H') {
				  binary++;
				  buf[1] = '0';
			  }
		  }
    }

		/* check any trailing chars */
		c = *(p-1);
		if( !hex && (c=='b' || c=='B') )
			{ binary++; *(p-1) = '\0'; }
		else if( c=='H' || c=='h' )
			{ hex++; *(p-1) = '\0'; }
		else if( !hex && (c=='D' || c=='d') )
			{ decimal++; *(p-1) = '\0'; }
		else if( c=='O' || c=='o' )
			{ octal++; *(p-1) = '\0'; }
		else if( !hex && !octal && !binary) decimal++;

		if (binary) {
			if (hex) error("ambiguous number, bin or hex");
			if (decimal) error("ambiguous number, bin or dec");
			if (octal) error("ambiguous number, bin or oct");
			for(p=buf; *p; p++ ) {
				if (*p == '1') {
          value = (value<<1) + 1;
        } else if (*p == '0') {
          value = value<<1;
        }	else {
				  error("Invalid binary digit: %c",*p);
        }
			}
			yylval.value = value;
			return(VALUE);
		}

		if (hex) {
			if (binary) error("ambiguous number, hex or bin");
			if (decimal) error("ambiguous number, hex or dec");
			if (octal) error("ambiguous number, hex or oct");
			for(p=buf; *p; p++ ) {
				value <<= 4;
				if( isdigit(*p) )
					value += *p-'0';
				else if( *p>='a' && *p<='f' )
					value += *p-'a'+ 10;
				else if( *p>='A' && *p<='F' )
					value += *p-'A'+ 10;
				else
				  error("Invalid hex digit: %c",*p);
			}
			yylval.value = value;
			return(VALUE);
		}

		if (decimal) {
			if (hex) error("ambiguous number, dec or hex");
			if (binary) error("ambiguous number, dec or bin");
			if (octal) error("ambiguous number, dec or oct");
			for(p=buf; *p; p++ ) {
				if( isdigit(*p) )
					value = value*10 + *p-'0';
				else
				   error("Invalid decimal digit: %c",*p);
			}
			yylval.value = value;
//      printf("Read decimal value %d\n", value);
			return(VALUE);
		}

		if (octal) {
			if (hex) error("ambiguous number, oct or hex");
			if (binary) error("ambiguous number, oct or bin");
			if (decimal) error("ambiguous number, oct or dec");
			for(p=buf; *p; p++ ) {
				if( *p>='0' && *p<='7' )
					value = (value << 3) + (*p - '0');
				else
				   error("Invalid octal digit: %c",*p);
			}
			yylval.value = value;
			return(VALUE);
		}

	default:
		if( isalpha(c) || c=='_' ) {
			p = buf;
			do {
				if( p-buf<sizeof(buf)-1 )
					*p++ = c;
				c = get_ch();
			} while( isalnum(c) || c=='_' );
			*p = '\0';
			unget_ch(c);
			if ( (op = lookop(buf)) != NULL ) {
				yylval.op = op;
				return(op->type);
			}
			sym = looksym(buf);
			yylval.sym = sym;
			return(SYMBOL);
		} else
			return(c);
	} /* switch */
} /* for */

} /* yylex */
