/* ----------------------------------------------------------------------
 * FILE: main.c
 * PACKAGE: as504 - Siemens SAB-C504 Assembler.
 *
 * DESCRIPTION:
 *	The file contains main(). It handles the arguments and makes
 *	sure that pass 1 is done before pass 2 etc...
 *
 * LICENCE:
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 * REVISION HISTORY:
 *	Jan. 19, 1990 - Created. (Ken Stauffer)
 *  Mar. 23, 2005 - Added -d debug option, "Press enter to continue" for
 *                  Windows, case insensitive checkext() and some changes
 *                  for working with includefiles (Freek van Walderveen).
 *
 * AUTHOR:
 *	Original file written by Ken Stauffer (University of Calgary).
 *	January, 1990. 
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <setjmp.h>

void emitusage(); // in emitter.c
void emitclose(); // in emitter.c
void emitaddr(unsigned int a); // in emitter.c
void emitopen(char *file, char *ftype, char *arg); // in emitter.c
void syminit();   // in symbol.c
void yyparse();   // in as504.{y,c}

extern int lineno;
extern int pass, fatal;
extern unsigned long lc;
extern FILE *fin;
extern int curregset;
extern int did_org;

jmp_buf main_env;
char *asmfile;
int dashl = 0;
FILE *listing;
int debug = 0;

/* ----------------------------------------------------------------------
 * checkext:
 *	Check the string s, for the presence of an extenstion e (case insensitive).
 *	Return the position of the start of e in s.
 *	or return NULL.
 */

char *checkext(char *s, char *e)
{
	register char *ps = s, *pe = e;

	while( *ps ) ps++;
	while( *pe ) pe++;

	for( ; ps>=s && pe>=e && (*ps|040) == (*pe|040); ps--, pe-- )
		if( pe == e ) return(ps);
	return(NULL);
}

void usage(char *cmd) {
  fprintf(stderr,"Usage: %s [-d] [-l] [-Ffmt] [-Aarg] infile.asm\n", cmd);
  emitusage();
}

int main(int argc, char *argv[])
{
	char *p,*dashF=NULL, *dashA=NULL;
	char objfile[100];
	char lstfile[100];
	int i;

	if( argc < 2 ) {
    usage(argv[0]);
#ifdef WIN32
    printf("\nPress enter to continue.\n");
    while (getchar() != '\n');
#endif
		exit(1);
	}

	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-' ) {
      break;
    }
		if (argv[i][1] == 'l' ) {
      dashl = 1;
    } else if (dashF == NULL && argv[i][1] == 'F') {
			dashF = argv[i]+2;
    } else if (dashA == NULL && argv[i][1] == 'A') {
			dashA = argv[i]+2;
    } else if (argv[i][1] == 'd') {
			debug = 1;
    } else {
			fprintf(stderr,"Duplicate or unknown flag.\n");
      usage(argv[0]);
#ifdef WIN32
      printf("\nPress enter to continue.\n");
      while (getchar() != '\n');
#endif
			exit(1);
		}
	}
	if( i == argc ) {
		fprintf(stderr,"Missing input file.\n");
#ifdef WIN32
    printf("\nPress enter to continue.\n");
    while (getchar() != '\n');
#endif
		exit(1);
	}

	if( (p=checkext(argv[i],".asm")) == NULL ) {
		fprintf(stderr,"Input file \"%s\" must end with .asm\n",
				argv[i]);
#ifdef WIN32
    printf("\nPress enter to continue.\n");
    while (getchar() != '\n');
#endif
		exit(1);
	}
  
  /* I want to be able to change asmfile later on */
  (void)strcpy((asmfile = (char *)malloc((unsigned)strlen(argv[i]) + 1)), argv[i]);
	/* asmfile = argv[i]; */

	if( (fin = fopen(argv[i],"r")) == NULL ) {
		fprintf(stderr,"Cannot open input file: %s\n",argv[i]);
#ifdef WIN32
    printf("\nPress enter to continue.\n");
    while (getchar() != '\n');
#endif
		exit(1);
	}

	if( dashl ) {
		strcpy(lstfile,argv[i]);
		strcpy(lstfile+(p-argv[i]),".lst");
		listing = fopen(lstfile,"w");
		if( listing == NULL ) {
			fprintf(stderr,"Cannot open file: %s for writting.\n",
				lstfile);
#ifdef WIN32
      printf("\nPress enter to continue.\n");
      while (getchar() != '\n');
#endif
			exit(1);
		}
	}
	strcpy(objfile,argv[i]);
	strcpy(objfile+(p-argv[i]),".hex");

	emitopen(objfile,dashF,dashA);

  syminit();
	fatal = 0;
	lineno = 1;
	pass=0;
	lc = 0x0000;

	if( setjmp(main_env) ) {
    if (debug) {
      printf("Exiting...\n");
    }
    if (fin != NULL) {
  		fclose(fin);
    }
		emitclose();
		unlink(objfile);
#ifdef WIN32
    printf("\nPress enter to continue.\n");
    while (getchar() != '\n');
#endif
		exit(1);
	}
  
	/*
	** P A S S    1
	*/
  if (debug) {
    printf("Pass 1...\n");
  }
	yyparse();
	if( fatal ) longjmp(main_env,1);
  
  if (!did_org) {
    fprintf(stderr,"File: %s, warning: You didn't use the ORG directive. This doesn't have to be a problem as long as you realise that the location counter is initially zero and you shouldn't assemble an included file on it's own as it will be missing initialisation!\n",asmfile);
  }
  
	rewind(fin);
	lineno = 1;
	pass++;
	lc = 0x0000;
	emitaddr(lc);
  curregset = 0;

	/*
	** P A S S    2
	*/
  if (debug) {
    printf("Pass 2...\n");
  }
	yyparse();

	emitclose();
	fclose(fin);
	if( dashl )
		fclose(listing);
  
  if (debug) {
    printf("Done\n");
  }
  
#ifdef WIN32
  printf("Assembling done, no errors. Press enter to continue.\n");
  while (getchar() != '\n');
#endif

  exit(0);
}
